//data.java Q1 2004
//Features: (ToDo*)
//  maintains list of players
//  maintains list of zones
//  maintains list of squads*
//  maintains list of banned players*
//  maintains list of ban free players*
//  maintains list of operators
//  hashes passwords to prevent malicous ops stealing them
//ToDo: Create a separate squad.dat listing owner and members

import java.io.*;
import java.util.*;
//import java.util.Hashtable;
//import java.util.Enumeration;
import java.util.zip.CRC32;

import java.util.Timer;
import java.util.TimerTask;

public class data
{
	static final int version = 7;

	//Modules
	static loggers lm;
	static dbio    io;

	static Hashtable users;
	static Hashtable online;
	static Hashtable zones;    //key = name, value = z (z.port)
	
	static Hashtable operators;
	static Hashtable squads;

	//static Timer persistTimer;
	static int  lastid;

	static java.text.SimpleDateFormat usageformat;
	
	public data(loggers lm, config cfg)
	{
		this.lm = lm;
		this.io = new dbio();
		
		users     = new Hashtable();
		online    = new Hashtable();
		zones     = new Hashtable();
		
		operators = new Hashtable();
		squads    = new Hashtable();
		lastid    = 0;
		
		usageformat = new java.text.SimpleDateFormat(
			"d-M-yyyy HH:mm:ss",
			new java.util.Locale("en","US").getDefault()
			);

		//Set up a timer callback for persist
		Timer persistTimer = new Timer(true); //daemon
		int seconds = cfg.GetInt("Persist", "SyncSeconds", 300) * 1000;
		persistTimer.schedule(new PersistTimer(this), seconds, seconds);

		//lm.LogM("D", "data", "Loaded v" + version);
	}

//// Misc util ////

	protected String PassToHash(String password)
	{
		CRC32 crc = new CRC32();
		crc.reset();
		crc.update( password.getBytes() );
		long passwd = crc.getValue();
		return "" + passwd;
	}

	//Seconds
	protected long getTime()
	{
		return new java.util.Date().getTime() / 1000;
	}

//// Player security ////

	public boolean existPlayer(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		return (p != null);
	}

	public boolean authPlayer(String name, String password)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return false;
		return p.password.equals( PassToHash(password) );
	}

	public boolean authOperator(String name)
	{
		Player p = (Player)operators.get(name.toLowerCase());
		return (p != null);
	}

	public boolean authSquadOwner(String name)
	{
		Player p = (Player)squads.get(name.toLowerCase());
		return (p != null);
	}

//// Player manipulation ////

	public void createPlayer(String name, String password)
	{
		Player p = new Player();
		p.name      = name;
		
		p.password  = PassToHash(password);
		p.firstused = getTime();
		p.usage     = 0;
		p.billerid  = "" + lastid; lastid++;
		users.put(name.toLowerCase(), p);
	}
/* Use getPorts() != null instead
	public boolean isOnline(String name)
	{
		Player p = (Player)online.get(name.toLowerCase());
		return (p != null);
	}
*/
	public void setOnline(String name, String pid, String port)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null)
		{
			lm.LogM("E", "data", "[" + name + "] cannot " +
				"set online status");
			return;
		}

		String moreThanOnce = "";
		p.pid.put(port, pid);
		if (online.put(name.toLowerCase(), p) == null)
		{
			p.start = getTime();

		}
		else moreThanOnce = " more than once";
		
		lm.LogM("I", "data", "{" + getZoneByPort(port) + "} [" +
			name + "] logged in" + moreThanOnce);
	}

	public void setOffline(String name, String port)
	{
		Player p = (Player)online.get(name.toLowerCase());
		if (p == null)
		{
			lm.LogM("E", "data", "[" + name + "] cannot " +
				"set offline status");
			return;
		}

		String completely = "";
		
		//Remove pid associated with port
		p.pid.remove(port);
		
		//Only remove from online list if...
		if (p.pid.size() == 0)
		{
			online.remove(p.name.toLowerCase());

			//Update usage
			long now = getTime();
			p.usage = p.usage + (now - p.start);

			completely = " completely";
		}
		
		lm.LogM("I", "data", "{" + getZoneByPort(port) + "} [" +
			name + "] logged out" + completely);
	}

	public void setAllOffline(String port_)
	{
		Enumeration all = online.elements();
		Enumeration ports;
		Player p = null;
		String port;
		
		//Enum all online players
		while(all.hasMoreElements())
		{
			p = (Player)all.nextElement();
			
			//Enum all ports
			ports = p.pid.keys();
			while(ports.hasMoreElements())
			{
				port = (String)ports.nextElement();
				if (port.equals(port_))
					setOffline(p.name, port);
			}
		}
	}

	public void setAllOffline()
	{
		Enumeration all = online.elements();
		Enumeration ports;
		Player p = null;
		String port;
		
		//Enum all online players
		while(all.hasMoreElements())
		{
			p = (Player)all.nextElement();
			
			//Enum all ports
			ports = p.pid.keys();
			while(ports.hasMoreElements())
			{
				port = (String)ports.nextElement();
				setOffline(p.name, port);
			}
		}
	}

	public void setBanner(String name, String banner)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return;
		p.banner = banner;
	}

	public String getBanner(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return null;
		return p.banner; //Could be null
	}

	public void setPassword(String name, String password)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return;
		p.password = PassToHash(password);
	}

//// Zone security ////

	public boolean existZone(String name)
	{
		Zone z = (Zone)zones.get(name.toLowerCase());
		return (z != null);
	}

	public boolean authZone(String name, String password)
	{
		Zone z = (Zone)zones.get(name.toLowerCase());
		if (z == null) return false;
		return z.password.equals( PassToHash(password) );
	}

//// Zone manipulation ////

	public void createZone(String name, String password)
	{
		Zone z = new Zone();
		z.name      = name;
		z.password  = PassToHash(password);
		zones.put(name.toLowerCase(), z);
	}
	
	public void zoneOnline(String network, String name, String port)
	{
		Zone z = (Zone)zones.get(name.toLowerCase());
		if (z == null) return;
		
		//Set all players inside the zone offline?
		
		z.network = network;
		z.port    = port;
	}

	//Need to figure out when/how to call this
	public void zoneOffline(String name, String port)
	{
		Zone z = (Zone)zones.get(name.toLowerCase());
		if (z == null) return;
		
		z.network = null;
		z.port    = null;
	}

//// Util IO ////

	public String getPidByName(String port, String name)
	{
		Player p = (Player)online.get(name.toLowerCase());
		if (p == null) return null;
		return (String)p.pid.get(port);
	}

	//Nooo...this operation is CPU intensive
	//Maybe keep a player list for each port, nested hashtables is confusing
	public String getNameByPid(String port, String pid)
	{
		Player p = null;
		String str = null;
		Enumeration all = online.elements();
		Enumeration ports;

		//Enum all online players
		while(all.hasMoreElements())
		{
			p = (Player)all.nextElement();
			
			//Enum all ports
			ports = p.pid.keys();
			while(ports.hasMoreElements())
			{
				str = (String)ports.nextElement();
				if (str.equals(port))
				{
					//Does the pid match?
					str = (String)p.pid.get(str);
					if (str.equals(pid))
						return p.name;
				}
			}
		}
		return "*Unknown*";
	}

	public String getName(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return "";
		return p.name;
	}

	public String getSquad(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return "";
		return p.squad;
	}
	
	public String getBillerID(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return null;
		return p.billerid;
	}
	
	public String getUsage(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return null;
		return "" + p.usage;
	}
	
	public String getFirstUsed(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return "";
		return usageformat.format(
			new java.util.Date(p.firstused*1000)
			);
	}

	public String getZoneByPort(String port)
	{
		//Search zones for a port match
		Zone z = null;
		Enumeration e = zones.elements();
		while (e.hasMoreElements())
		{
			z = (Zone)e.nextElement();
			if (port.equals(z.port))
			{
				if (z.network.length () > 0)
					return z.network + " " + z.name;
				return z.name;
			}
		}
	
		return null;
	}

	public String getZone(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return null;
		
		p = (Player)online.get(name.toLowerCase());
		if (p == null) return "Not online.";
		
		//Search logins for any port
		String port = null;
		Enumeration e = p.pid.keys();
		while (e.hasMoreElements())
		{
			port = (String)e.nextElement();
		}
		
		return getZoneByPort(port);
	}

	public Enumeration getPorts(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return null;
		return p.pid.keys();
	}

	public Enumeration getPids(String name)
	{
		Player p = (Player)users.get(name.toLowerCase());
		if (p == null) return null;
		return p.pid.elements();
	}

//// File IO ////

	public void load()
	{
		loadUsers("user.dat");
		loadZones("zone.dat");
		loadOperators("operator.dat");
	}

	public void flush()
	{
		flushUsers("user.dat");
		flushZones("zone.dat");
		flushOperators("operator.dat");
	}

	//user.dat
	public void loadUsers(String fname)
	{
		//lm.LogM("D", "data", "Reading " + fname);
		BufferedReader in = io.fopenr(fname);
		if (in == null) return;

		lastid = -1;
		int usercount = 0;
		Player p = null;

		String[] part;
		String key, value;
		while(true)
		{
			part = io.readLine(in);
			if (part == null) break;
			key   = part[0];
			value = part[1];
			
			if (key.equalsIgnoreCase("NAME"))
			{
				p = (Player)users.get(value.toLowerCase());
				if (p == null)
				{
					p = new Player();
					users.put(value.toLowerCase(), p);
					usercount++;
				}
				p.name = value;
			}
			else if (key.equalsIgnoreCase("SQUAD"))
			{
				p.squad = value;
			}
			else if (key.equalsIgnoreCase("PASSWORD"))
			{
				p.password = value;
			}
			else if (key.equalsIgnoreCase("IP"))
			{
				p.ip.put( "" + p.ip.size(), value );
			}
			else if (key.equalsIgnoreCase("MACID"))
			{
				p.macid.put( "" + p.macid.size(), value );
			}
			else if (key.equalsIgnoreCase("CONTID"))
			{
				p.contid.put( "" + p.contid.size(), value );
			}
			else if (key.equalsIgnoreCase("BANNER"))
			{
				p.banner = value;
			}
			else if (key.equalsIgnoreCase("FIRSTUSED"))
			{
				p.firstused = Long.parseLong( value );
			}
			else if (key.equalsIgnoreCase("USAGE"))
			{
				p.usage = Long.parseLong( value );
			}
			else if (key.equalsIgnoreCase("BILLERID"))
			{
				int i = Integer.parseInt( value );
				if (i > lastid) lastid = i;
				p.billerid = value;
			}
		}
		
		io.close(in);
		
		lastid++;
		lm.LogM("I", "data", "user.dat ....... " + users.size() +
			" (" + lastid + ")");
		if (users.size() != lastid)
			lm.LogM("W", "data", "users.dat may be corrupt");
	}

	//user.dat
	public void flushUsers(String fname)
	{
		PrintWriter out = io.fopenw(fname);
		if (out == null) return;
		Player p;

		Enumeration e = users.elements();
		while(e.hasMoreElements())
		{
			p = (Player)e.nextElement();
			out.println("NAME:"     + p.name);
			if (p.squad != null && p.squad.length() > 0)
				out.println("SQUAD:"    + p.squad);
			out.println("PASSWORD:" + p.password);

			writeTable(out, p.ip,     "IP:");
			writeTable(out, p.macid,  "MACID:");
			writeTable(out, p.contid, "CONTID:");

			if (p.banner != null && p.banner.length() > 0)
				out.println("BANNER:" + p.banner);
			out.println("FIRSTUSED:" + p.firstused);
			out.println("USAGE:" + p.usage);
			out.println("BILLERID:" + p.billerid);
		}

		out.close();
	}

	//zone.dat
	public void loadZones(String fname)
	{
		//lm.LogM("D", "data", "Reading " + fname);
		BufferedReader in = io.fopenr(fname);
		if (in == null) return;
		
		Zone z = null;
		
		String[] part;
		String key, value;
		while(true)
		{
			part = io.readLine(in);
			if (part == null) break;
			key   = part[0];
			value = part[1];

			if (key.equalsIgnoreCase("NAME"))
			{
				z = (Zone)zones.get(value.toLowerCase());
				if (z == null)
				{
					z = new Zone();
					z.name = value;
					zones.put(value.toLowerCase(), z);
				}
			}
			else if (key.equalsIgnoreCase("PASSWORD"))
			{
				z.password = value;
			}
		}

		io.close(in);
		lm.LogM("I", "data", "zone.dat ....... " + zones.size());
	}

	//zone.dat
	public void flushZones(String fname)
	{
		PrintWriter out = io.fopenw(fname);
		if (out == null) return;
		Zone z;

		Enumeration e = zones.elements();
		while(e.hasMoreElements())
		{
			z = (Zone)e.nextElement();
			out.println("NAME:"     + z.name);
			out.println("PASSWORD:" + z.password);
		}

		out.close();
	}

	//operator.dat
	public void loadOperators(String fname)
	{
		//lm.LogM("D", "data", "Reading " + fname);
		BufferedReader in = io.fopenr(fname);
		if (in == null) return;
		
		Player p;
		
		String[] part;
		String key, value;
		while(true)
		{
			part = io.readLine(in);
			if (part == null) break;
			key   = part[0];
			value = part[1];

			if (key.equalsIgnoreCase("NAME"))
			{
				p = (Player)operators.get(value.toLowerCase());
				if (p == null)
				{
					p = (Player)users.get(value.toLowerCase());
					if (p != null)
						operators.put(value.toLowerCase(), p);
				}
			}
		}

		io.close(in);
		lm.LogM("I", "data", "operator.dat ... " + operators.size());
	}

	//operator.dat
	public void flushOperators(String fname)
	{
		PrintWriter out = io.fopenw(fname);
		if (out == null) return;
		Player p;

		Enumeration e = operators.elements();
		while(e.hasMoreElements())
		{
			p = (Player)e.nextElement();
			out.println("NAME:"     + p.name);
		}

		out.close();
	}

	private void writeTable(PrintWriter out,
		Hashtable ht, String prefix)
	{
		String str;
		Enumeration e = ht.elements();

		while(e.hasMoreElements())
		{
			str = (String)e.nextElement();
			out.println(prefix + str);
		}
	}

//// Persist ////

	class PersistTimer extends TimerTask
	{
		data pd;
	
		public PersistTimer(data pd)
		{
			this.pd = pd;
		}
	
		public void run()
		{
			//lm.LogM("D", "data", "Saving persistent data");
			pd.flush();
		}
	}

//// Containers ////

	class Player
	{
		static final int version = 1;

		public String    name;
		public String    squad;
		public String    password;
		public Hashtable ip;
		public Hashtable macid;
		public Hashtable contid;

		public String    banner;
		public long      firstused;
		public long      usage;
		public String    billerid; //userid

		public Hashtable pid;      //key = port, value = pid
		public Hashtable channels;
		public long      start;
		
		public Player()
		{
			name      = null;
			squad     = "";
			password  = null;
			ip        = new Hashtable();
			macid     = new Hashtable();
			contid    = new Hashtable();

			banner    = "";
			firstused = 0;
			usage     = 0;
			billerid  = "";

			//This stuff is not persistent
			pid       = new Hashtable();
			channels  = new Hashtable();
			start     = 0;
		}
	}

	class Zone
	{
		static final int version = 1;

		public String name;
		public String password;

		public String network;
		public String port;
		
		public Zone()
		{
			name     = null;
			password = null;
			
			//This stuff is not persistent
			network  = null;
			port     = null;
		}
	}
}
